#
# Copyright (C) 1997-1999 Sensus Consulting Ltd. All Rights Reserved.
# Matt Newman <matt@sensus.org> and Jean-Claude Wippler <jcw@equi4.com>
#
# $Header: /cvsroot/metakit/mk/tcl/kit/kitsh.vfs/lib/vfs/mk4vfs.tcl,v 1.2 2000/12/04 23:50:14 jcw Exp $
#

###############################################################################
# use Pink for zip and md5 replacements, this avoids the dependency on Trf

    package ifneeded Trf 1.3 {
    
        package require pink
        package provide Trf 1.3
    
        proc zip {flag value data} {
            switch -glob -- "$flag $value" {
            {-mode d*} {
                set mode decompress
            }
            {-mode c*} {
                set mode compress
            }
            default {
                error "usage: zip -mode {compress|decompress} data"
            }
            }
            return [pink zlib $mode $data]
        }
    
        proc crc {data} {
            return [pink zlib crc32 $data]
        }
    
        proc md5 {data} {
            set cmd [pink md5]
            $cmd update $data
            set result [$cmd digest]
            rename $cmd ""
            return $result
        }
    }

###############################################################################
# this replacement is for memchan, used for simple (de)compression

    package ifneeded memchan 0.1 {
    
        package require rechan
        package provide memchan 0.1
    
        proc _memchan_handler {cmd fd args} {
            upvar #0 ::_memchan_buf($fd) _buf
            upvar #0 ::_memchan_pos($fd) _pos
            set arg1 [lindex $args 0]
            
            switch $cmd {
                seek {
                    switch $args {
                        1 { incr arg1 $_pos }
                        2 { incr arg1 [string length $_buf]}
                    }
                    return [set _pos $arg1]
                }
                read {
                    set r [string range $_buf $_pos [expr { $_pos + $arg1 - 1 }]]
                    incr _pos [string length $r]
                    return $r
                }
                write {
                    set n [string length $arg1]
                    if { $_pos >= [string length $_buf] } {
                        append _buf $arg1
                    } else { # the following doesn't work yet :(
                        set last [expr { $_pos + $n - 1 }]
                        set _buf [string replace $_buf $_pos $last $arg1]
                        error "mk4vfs: sorry no inline write yet"
                    }
                    incr _pos $n
                    return $n
                }
                close {
                    unset _buf _pos
                }
                default {
                    error "Bad call to memchan replacement handler: $cmd"
                }
            }
        }
        
        proc memchan {} {
            set fd [rechan _memchan_handler 6]
            #fconfigure $fd -translation binary -encoding binary
            
            set ::_memchan_buf($fd) ""
            set ::_memchan_pos($fd) 0
            
            return $fd
        }
    }
        
###############################################################################

package require Mk4tcl
package require vfs

package provide mk4vfs 0.6

namespace eval mk4vfs {
    variable uid 0
    variable compress 1         ;# HACK - needs to be part of "Super-Block"
    variable flush      5000    ;# Auto-Commit frequency
    variable direct 0

    namespace export mount umount
}

proc mk4vfs::init {db} {
    mk::view layout $db.dirs \
            {name:S parent:I {files {name:S size:I date:I contents:M}}}

    if { [mk::view size $db.dirs] == 0 } {
        mk::row append $db.dirs name <root> parent 0
    }
}

proc mk4vfs::mount {path file args} {
    variable uid
    set db mk4vfs[incr uid]

    eval [list mk::file open $db $file] $args

    init $db

    ::vfs::mount $path mk4vfs::driver $db

    set flush 1
    for {set idx 0} {$idx < [llength $args]} {incr idx} {
        switch -- [lindex $args $idx] {
        -readonly       -
        -nocommit       {set flush 0}
        }
    }
    if { $flush } {
        _commit $db
    }
    return $db
}

proc mk4vfs::_commit {db} {
    after ${::mk4vfs::flush} [list mk4vfs::_commit $db]
    mk::file commit $db
}

proc mk4vfs::umount {path args} {
    return [eval [list vfs::umount $path] $args]
}

proc mk4vfs::stat {db path arr} {
    variable cache
    upvar 1 $arr sb
    #tclLog "mk4vfs::stat $db $path $arr"

    set sp [::tcl::file split $path]
    set tail [lindex $sp end]

    set parent 0
    set view $db.dirs
    set cur $view!$parent
    set type directory

    foreach ele [lrange $sp 0 [expr { [llength $sp] - 2 }]] {

        if { [info exists cache($cur,$ele)] } {
            set parent $cache($cur,$ele)
        } else {
            #set row [mk::select $view name $ele parent $parent]
            set row [find/dir $view $ele $parent]

            if { $row == -1 } {
                #tclLog "select failed: parent $parent name $ele"
                return -code error "could not read \"$path\": no such file or directory"
            }
        set parent $row
            set cache($cur,$ele) $parent
        }
    set cur $view!$parent
    #mk::cursor position cur $parent
    }
    #
    # Now check if final comp is a directory or a file
    #
    # CACHING is required - it can deliver a x15 speed-up!
    #
    if { [string equal $tail "."] || [string equal $tail ":"] || \
                        [string equal $tail ""] } {
    # donothing

    } elseif { [info exists cache($cur,$tail)] } {
        set type directory
        #set cur $view!$cache($cur,$tail)
    mk::cursor position cur $cache($cur,$tail)

    } else {
        # File?
        #set row [mk::select $cur.files name $tail]
        set row [find/file $cur.files $tail]

        if { $row != -1 } {
            set type file
            set view $cur.files
        #set cur $view!$row
        mk::cursor create cur $view $row

        } else {
            # Directory?
            #set row [mk::select $view parent $parent name $tail]
            set row [find/dir $view $tail $parent]

            if { $row != -1 } {
                set type directory
        #set cur $view!$row
        mk::cursor position cur $row
                set cache($cur,$tail) $row
            } else { 
                return -code error "could not read \"$path\": no such file or directory"
            }
        }
    }
    set sb(type)    $type
    set sb(view)    $view
    set sb(ino)     $cur
    set sb(dev)     [list mk4vfs::driver $db]

    if { [string equal $type "directory"] } {
        set sb(atime)   0
        set sb(ctime)   0
    set sb(gid) 0
        set sb(mode)    0777
        set sb(mtime)   0
        set sb(nlink)   [expr { [mk::get $cur files] + 1 }]
        set sb(size)    0
        set sb(csize)   0
    set sb(uid) 0
    } else {
        set mtime   [mk::get $cur date]
        set sb(atime)   $mtime
        set sb(ctime)   $mtime
    set sb(gid) 0
        set sb(mode)    0777
        set sb(mtime)   $mtime
        set sb(nlink)   1
        set sb(size)    [mk::get $cur size]
        set sb(csize)   [mk::get $cur -size contents]
    set sb(uid) 0
    }
}

proc mk4vfs::driver {db option args} {
    #tclLog "mk4vfs::driver $db $option $args"
    switch -- $option {
    lstat       {return [uplevel 1 [concat [list mk4vfs::stat $db] $args]]}
    chdir       {return [lindex $args 0]}
    mtime       -
    delete      -
    stat        -
    getdir      -
    mkdir       {return [uplevel 1 [concat [list mk4vfs::$option $db] $args]]}
    
    open        {
            set file [lindex $args 0]
            set mode [lindex $args 1]

            switch -glob -- $mode {
            {}  -
            r   {
                    stat $db $file sb
                
                    if { $sb(csize) != $sb(size) } {
                        package require Trf
                        package require memchan
                        #tclLog "$file: decompressing on read"

                        set fd [memchan]
                        fconfigure $fd -translation binary
                        set s [mk::get $sb(ino) contents]
                        puts -nonewline $fd [zip -mode decompress $s]

                        fconfigure $fd -translation auto
                        seek $fd 0
			@close $fd [list _memchan_handler close $fd]
                    } elseif { $::mk4vfs::direct } {
                        package require Trf
                        package require memchan

                        set fd [memchan]
                        fconfigure $fd -translation binary
                        puts -nonewline $fd [mk::get $sb(ino) contents]

                        fconfigure $fd -translation auto
                        seek $fd 0
			@close $fd [list _memchan_handler close $fd]
		    } else {
			set fd [mk::channel $sb(ino) contents r]
                    }
                    return $fd
                }
            a   {
                    if { [catch {stat $db $file sb }] } {
                        #tclLog "stat failed - creating $file"
                        # Create file
                        stat $db [file dirname $file] sb

                        set cur [mk::row append $sb(ino).files \
                            name [file tail $file] size 0 date [clock seconds] ]
                        set sb(ino) $cur

                        if { [string match *z* $mode] || ${mk4vfs::compress} } {
                            set sb(csize) -1    ;# HACK - force compression
                        } else {
                            set sb(csize) 0
                        }
                    }

                    if { $sb(csize) != $sb(size) } {
                        package require Trf
                        package require memchan

                        #tclLog "$file: compressing on append"
                        append mode z
                        set fd [memchan]

                        fconfigure $fd -translation binary
                        set s [mk::get $sb(ino) contents]
                        puts -nonewline $fd [zip -mode decompress $s]
                        fconfigure $fd -translation auto
                    } else {
            set fd [mk::channel $sb(ino) contents a]
                    }
                    @close $fd [list mk4vfs::do_close $fd $mode $sb(ino)]
                    return $fd
                }
            w*  {
                    if { [catch {stat $db $file sb }] } {
                        #tclLog "stat failed - creating $file"
                        # Create file
                        stat $db [file dirname $file] sb
                        set cur [mk::row append $sb(ino).files \
                            name [file tail $file] size 0 date [clock seconds] ]
                        set sb(ino) $cur
                    }
                    if { [string match *z* $mode] || ${mk4vfs::compress} } {
                        package require Trf
                        package require memchan
                        #tclLog "$file: compressing on write"
                        ###zip -attach $fd -mode compress
                        append mode z
                        set fd [memchan]
                    } else {
                        set fd [mk::channel $sb(ino) contents w]
                    }
                    @close $fd [list mk4vfs::do_close $fd $mode $sb(ino)]
                    return $fd
                }
            default     {
                    error "illegal access mode \"$mode\""
                }
            }
        }
    sync        {eval [list mk::file commit $db] [lrange $args 1 end]}
    umount      {eval [list mk::file close $db] $args}
    default     {
            return -code error "bad option \"$option\": must be one of chdir, delete, getdir, load, lstat, mkdir, open, stat, sync, or umount"
        }
    }
}

proc mk4vfs::do_close {fd mode cur} {
    # Set size to -1 before the seek - just in case it fails.
    mk::set $cur size -1 date [clock seconds]

    flush $fd
    if { [string match *z* $mode] } {
        #puts [fconfigure $fd]
        fconfigure $fd -translation binary
        ###unstack $fd
        #puts [fconfigure $fd]
        fconfigure $fd -translation binary
        seek $fd 0
        set data [read $fd]
        set cdata [zip -mode compress $data]
        set len [string length $data]
        set clen [string length $cdata]
        if { $clen < $len } {
            mk::set $cur size $len contents $cdata
        } else {
            mk::set $cur size $len contents $data
        }
    } else {
        mk::set $cur size [mk::get $cur -size contents]
    }

    # added 30-10-2000
    set db [lindex [split $cur .] 0]
    mk::file autocommit $db
}

proc mk4vfs::mkdir {db path} {
    set sp [::tcl::file split $path]
    set parent 0
    set view $db.dirs

    set npath {}
    foreach ele $sp {
        set npath [file join $npath $ele]

        if { ![catch {stat $db $npath sb}] } {
            if { $sb(type) != "directory" } {
                return -code error "can't create directory \"$npath\": file already exists"
            }
            set parent [mk::cursor position sb(ino)]
            continue
        }
        #set parent [mk::cursor position sb(ino)]
#puts "set cur \[mk::row append $view name $ele parent $parent]"
        set cur [mk::row append $view name $ele parent $parent]
        set parent [mk::cursor position cur]
    }
}

proc mk4vfs::getdir {db path {pat *}} {
    #tclLog [list mk4vfs::getdir $db $path $pat]

    if { [catch {
        stat $db $path sb
    }] } {
        return {}
    }

    if { $sb(type) != "directory" } {
        return {}
        #return -code error "bad path \"$path\": not a directory"
    }
    # Match directories
    set parent [mk::cursor position sb(ino)] 

if { 0 } {
    foreach row [mk::select $sb(view) parent $parent -glob name $pat] {
        if { $row == 0 } {continue}

        set hits([mk::get $sb(view)!$row name]) 1
    }
    # Match files
    set view $sb(view)!$parent.files
    foreach row [mk::select $view -glob name $pat] {
        set hits([mk::get $view!$row name]) 1
    }
} else {
    mk::loop sb(ino) {
    if { [mk::get $sb(ino) parent] == $parent &&
         [string match $pat [mk::get $sb(ino) name]] &&
         [mk::cursor position sb(ino)] != 0 } {
        set hits([mk::get $sb(ino) name]) 1
    }
    }
    # Match files
    mk::loop sb(ino) $sb(view)!$parent.files {
    if { [string match $pat [mk::get $sb(ino) name]] } {
        set hits([mk::get $sb(ino) name]) 1
    }
    }
}
    return [lsort [array names hits]]
}

proc mk4vfs::mtime {db path time} {

    stat $db $path sb

    if { $sb(type) == "file" } {
        mk::set $sb(ino) date $time
    }
    return $time
}

proc mk4vfs::delete {db args} {

    set force 0
    while {[llength $args] > 0} {
        set arg [lindex $args 0]
        if { $arg == "-force" } {
            set force 1
            set args [lrange $args 1 end]
        } elseif { $arg == "--" } {
            set args [lrange $args 1 end]
            break
        } else {
            break
        }
    }
    foreach path $args {
        set rc [catch { stat $db $path sb } err]
        if { $rc }  {
            if { $force } {
                continue
            } else {
                return -code error $err
            }
        }

        if { $sb(type) == "file" } {
            mk::row delete $sb(ino)
        } else {
            # just mark dirs as deleted
            set tail [file tail $path]
            catch {unset cache($sb(ino),$tail)}
            mk::set $sb(ino) parent -1
        }
    }
    return ""
}

proc mk4vfs::find/file {v name} {
    mk::loop cur $v {
    if { [string equal [mk::get $cur name] $name] } {
        return [mk::cursor position cur]
    }
    }
    return -1
}

proc mk4vfs::find/dir {v name parent} {
    mk::loop cur $v {
    if {    [mk::get $cur parent] == $parent &&
        [string equal [mk::get $cur name] $name] } {
        return [mk::cursor position cur]
    }
    }
    return -1
}
