/*
 * Matt Newman <matt@sensus.org>
 * Changes by Jean-Claude Wippler <jcw@equi4.com> (with the help of Vince).
 *
 * $Header: /cvsroot/metakit/mk/tcl/kit/pwb.c,v 1.1 2000/11/08 22:47:57 jcw Exp $
 *
 * Programmers Workbench....
 */

#include <tcl.h>
#include <tclInt.h> /* TclGetLibraryPath */
#include <string.h>

/*
 * Exported Routines
 */

DLLEXPORT Tcl_AppInitProc Pwb_Init;

/*
 * Internal Routines
 */

static Tcl_ObjCmdProc	DumpObjCmd;
static Tcl_ObjCmdProc	LConcatObjCmd;
static Tcl_ObjCmdProc	NopObjCmd;
static Tcl_ObjCmdProc	LibraryPathObjCmd;

/*
 * Helper Defines
 */
#define u_char	unsigned char
#define u_int	unsigned int

int
DumpObjCmd(dummy, interp, objc, objv)
	ClientData dummy;
	Tcl_Interp *interp;
	int objc;
	Tcl_Obj *CONST objv[];
{
    static char hex[] = "0123456789abcdef";
    Tcl_Obj *objPtr;
    u_char *scp, *dcp;
    u_char c1 = '\0';
    char *prefix = NULL;
    int sclen, dclen, plen = 0, i, idx, count;
    u_int val;

    if (objc != 2 && objc != 3) {
	Tcl_WrongNumArgs(interp, 1, objv, "data ?prefix?");
	return TCL_ERROR;
    }

    scp = Tcl_GetByteArrayFromObj( objv[1], &sclen);

    if (objc > 2)
	prefix = Tcl_GetStringFromObj( objv[2], &plen);

    /*
     * Max line length = (64 +strlen(prefix))
     * and 16 bytes output per line...
     */
    objPtr = Tcl_NewByteArrayObj( &c1, 0);
    Tcl_SetByteArrayLength( objPtr, (64+plen) * (1+sclen/16) + 1);
    dcp = Tcl_GetByteArrayFromObj( objPtr, NULL);

    dclen = idx =  0;
    while (idx < sclen) {
	if (idx > 0)
	    dcp[dclen++] = '\n';

	/* Output Line Prefix */
	if (plen) {
	    strncpy( (char*) &dcp[dclen], prefix, plen);
	    dclen += plen;
	}

	/* Output Offset in Hex */
	val = idx;
	for (i = 32; i > 0;i -= 4) {
	    dcp[dclen++] = hex[ (val >> (i-4)) & 0xf ];
	}
	dcp[dclen++] = ' ';

	/* Output 4 data blocks of 4 bytes hex */
	count = idx;
	for (i = 1; i <= 16; i++,count++) {
	    if ( count >= sclen ) {
		dcp[dclen++] = ' ';
		dcp[dclen++] = ' ';
	    } else {
		dcp[dclen++] = hex[ (scp[count] >> 4) & 0xf];
		dcp[dclen++] = hex[ scp[count] & 0xf];
	    }

	    if ( (i % 4) == 0 )
		dcp[dclen++] = ' ';
	}
	dcp[dclen++] = '*';

	/* Output 16 bytes as printable ASCI */
	count = idx;
	for (i = 1; i <= 16; i++,count++) {
	    if ( count >= sclen )
		dcp[dclen++] = ' ';
	    else if ( scp[count] < 32 || scp[count] > 126 )
		dcp[dclen++] = '.';
	    else
		dcp[dclen++] = scp[count];
	}
	dcp[dclen++] = '*';

	idx += 16;
    }
    Tcl_SetByteArrayLength( objPtr, dclen);
    Tcl_SetObjResult( interp, objPtr);
    return TCL_OK;
}

int
LConcatObjCmd(dummy, interp, objc, objv)
	ClientData dummy;
	Tcl_Interp *interp;
	int objc;
	Tcl_Obj *CONST objv[];
{
    Tcl_Obj *listPtr;
    int idx;

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "list ?list ...?");
	return TCL_ERROR;
    }

    listPtr = Tcl_NewListObj( 0, (Tcl_Obj **)NULL);

    for (idx = 1; idx < objc; idx++) {
	if (Tcl_ListObjAppendList(interp, listPtr, objv[idx]) != TCL_OK) {
	    Tcl_DecrRefCount( listPtr);
	    return TCL_ERROR;
	}
    }
    Tcl_SetObjResult( interp, listPtr);
    return TCL_OK;
}

/* Does nothing - for timing purposes */
int
NopObjCmd(dummy, interp, objc, objv)
	ClientData dummy;
	Tcl_Interp *interp;
	int objc;
	Tcl_Obj *CONST objv[];
{
    return TCL_OK;
}

/* Support for encodings, from Vince Darley <vince.darley@eurobios.com> */
int
LibraryPathObjCmd(dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
     if (objc == 1) {
	Tcl_SetObjResult(interp, TclGetLibraryPath());
     } else {
	TclSetLibraryPath(objv[1]);
     }
     return TCL_OK;
}

/*
 * Public Entrypoint
 */

DLLEXPORT int Pwb_Init(Tcl_Interp *interp)
{
    Tcl_CreateObjCommand(interp, "hexdump", DumpObjCmd, 0, 0);
    Tcl_CreateObjCommand(interp, "lconcat", LConcatObjCmd, 0, 0);
    Tcl_CreateObjCommand(interp, "nop", NopObjCmd, 0, 0);
    Tcl_CreateObjCommand(interp, "librarypath", LibraryPathObjCmd, 0, 0);

    return Tcl_PkgProvide( interp, "pwb", "0.2");
}
