proc tclInit {} {
    global tcl_library tcl_version tcl_libPath auto_path tcl_root tcl_platform
    global argc argv0 argv
    
    set tcl_platform(vfs) 1
    set tcl_root [info nameofexecutable]
    set tcl_library [file join $tcl_root lib tcl$tcl_version]
    set tcl_libPath [list $tcl_library [file join $tcl_root lib]]

# hack alert: this lets older scripted docs work with newer Mk4tcl versions

    rename package _tclkit_package
    
    proc package {args} {
        set a [join $args]
        if {[string match "require Mk4tcl 1.*" $a]} {
            set args "require Mk4tcl"
        } elseif {[string match "require Trf 1.*" $a]} {
            set args "require Trf"
        }
        eval _tclkit_package $args
    }

# use Pink for zip and md5 replacements, this avoids the dependency on Trf

    package ifneeded Trf 1.3 {
    
        package require pink
        package provide Trf 1.3
    
        proc zip {flag value data} {
            switch -glob -- "$flag $value" {
            {-mode d*} {
                set mode decompress
            }
            {-mode c*} {
                set mode compress
            }
            default {
                error "usage: zip -mode {compress|decompress} data"
            }
            }
            return [pink zlib $mode $data]
        }
    
        proc crc {data} {
            return [pink zlib crc32 $data]
        }
    
        proc md5 {data} {
            set cmd [pink md5]
            $cmd update $data
            set result [$cmd digest]
            rename $cmd ""
            return $result
        }
    }

# lookup /lib/vfs/{vfs.tcl,mk4vfs.tcl}
    load {} vfs

    set d [mk::select exe.dirs parent 0 name lib]
    set d [mk::select exe.dirs parent $d name vfs]
    
    set n [mk::select exe.dirs!$d.files name vfs.tcl]
    set s [mk::get exe.dirs!$d.files!$n contents]
    catch {set s [pink zlib decompress $s]}
    uplevel #0 $s
    
    set n [mk::select exe.dirs!$d.files name mk4vfs.tcl]
    set s [mk::get exe.dirs!$d.files!$n contents]
    catch {set s [pink zlib decompress $s]}
    uplevel #0 $s

    load {} pwb
    load {} rechan
    vfs::mount $tcl_root mk4vfs::driver exe

    # re-install VFS as a C-level intercept if vfschan exists (01-11-2000)
    catch {
        vfschan
	rename ::open vfs::open
	proc ::open {args} {uplevel 1 ::tcl::open $args}
    }

    # alter path to find encodings, if pwb defines the "librarypath" cmd
    catch {
	set path [librarypath]
	lappend path [info library]
	librarypath $path
    }

    uplevel #0 [list source [file join $tcl_library init.tcl]]
# reset auto_path
    set auto_path $tcl_libPath
    
    rename tclInit {}
}
